#include "JMap2PixelObj.h"
#include "JInteger.h"
#include "JIntegerData.h"
#include "JStringData.h"
#include "JRealData.h"
#include "JIntegerProperty.h"
#include "JIntegerListProperty.h"
#include "JStringProperty.h"
#include "JLinkObj.h"
#include "JBytesData.h"

#include <malloc.h>
#include <string.h>
#include <stdio.h>

static char* newstrcpy( char* s ) {
	char* result = (char*)malloc( strlen( s ) + 1 );
	strcpy( result, s );
	return( result );
}

char* theJMap2PixelObj = JMap2PixelObj().Register();


void JMap2PixelObj::writeContent(JOutputStream& os) {
  JAddObj::writeContent(os);
  putObject(os, "ul_latitude", ul_latitude);
  putObject(os, "ul_longitude", ul_longitude);
  putObject(os, "lr_latitude", lr_latitude);
  putObject(os, "lr_longitude", lr_longitude);
  putObject(os, "map_name", map_name);
  putInteger(os, "pixel_width", pixel_width);
  putInteger(os, "pixel_height", pixel_height);
}

void JMap2PixelObj::readContent(JDictionary& dict) {
  JAddObj::readContent(dict);
  JObject* obj_ul_latitude = getObject( dict, "ul_latitude");
  if ( obj_ul_latitude ) ul_latitude = *(JString*)obj_ul_latitude;
  JObject* obj_ul_longitude = getObject( dict, "ul_longitude");
  if ( obj_ul_longitude ) ul_longitude = *(JString*)obj_ul_longitude;
  JObject* obj_lr_latitude = getObject( dict, "lr_latitude");
  if ( obj_lr_latitude ) lr_latitude = *(JString*)obj_lr_latitude;
  JObject* obj_lr_longitude = getObject( dict, "lr_longitude");
  if ( obj_lr_longitude ) lr_longitude = *(JString*)obj_lr_longitude;
  JObject* obj_map_name = getObject( dict, "map_name");
  if ( obj_map_name ) map_name = *(JString*)obj_map_name;
  pixel_width = getInteger(dict, "pixel_width");
  pixel_height = getInteger(dict, "pixel_height");
  if ( pixel_width <= 0 ) {
	  pixel_width = 1;
  }
  if ( pixel_height <= 0 ) {
	  pixel_height = 1;
  }
  if (( ul_longitude.length() > 0 ) && ( lr_longitude.length() > 0 ) && 
	  ( ul_latitude.length() > 0 ) && ( lr_latitude.length() > 0 ) &&
	  ( pixel_width > 1 ) && ( pixel_height > 1 )) {
	pixelMapper.init( ul_longitude, lr_longitude, ul_latitude, lr_latitude, map_name,
	  pixel_width, pixel_height );
  }
  remap( longitude, latitude );
}

const char* JMap2PixelObj::className() const { return "JMap2PixelObj";}


JObject* JMap2PixelObj::clone() const { return new JMap2PixelObj(*this);}


// class constructor
JMap2PixelObj::JMap2PixelObj() {
  igm = 1 + 1 + 1 + 1 + 7;
  ogm = 1 + 1 + 1;
  pixel_width = 0;
  pixel_height = 0;
  pixel_x = 0;
  pixel_y = 0;
  in_map = 0;
  d_longitude = 0;
  d_latitude = 0;
  lastRemapType = 0;
}


void JMap2PixelObj::draw(JGraphics g, int x, int y, int w, int h) {
  JRect rect(x, y, w, h);
  if ( map_name.length() == 0 ) {
	  drawText(g, JString("MAP"), rect);
  } else {
	  drawText(g, map_name, rect);
  }
}


boolean JMap2PixelObj::inputAllowed(int n)
{ return !inputSet(n).last();}
JString JMap2PixelObj::inputTag(int n) {
  char* tag[] = {
    "latitude",
		"longitude", 
  "real_latitude",
  "real_longitude",
  "map name",
  "upper left latitude",
  "upper left longitude",
  "lower right latitude",
  "lower right longitude",
  "pixel width",
  "pixel height" };
  return( tag[n] );
}


int JMap2PixelObj::inputType( int n ) {
  if ( n == 0 ) return JStringData::id;
  else if ( n == 1 ) return JStringData::id;
  else if ( n == 2 ) return JRealData::id;
  else if ( n == 3 ) return JRealData::id;
  else if ( n == 4 ) return JStringData::id;
  else if (( n >= 5 ) && ( n <= 8 )) {
	  // accept any data type here (can only handle int, real, string though)
	  int type = JDataType::last;
	  JLinkObj* link = (JLinkObj*)inputSet( n ).first();
	  if ( link ) {
		  type = link->from->outputType( link->fn );
	  }
	  return( type );
  }
  /*if ( n == 5 ) return JRealData::id;
  if ( n == 6 ) return JRealData::id;
  if ( n == 7 ) return JRealData::id;
  if ( n == 8 ) return JRealData::id;*/
  else if ( n == 9 ) return JIntegerData::id;
  else if ( n == 10 ) return JIntegerData::id;
  else return JIntegerData::id;
}
JString JMap2PixelObj::outputTag(int n) {
  char* tag[] = {
    "in_map",
    "pixel_x",
    "pixel_y" };
  return( tag[n] );
}


JFRect JMap2PixelObj::inputArea(int n) {
  if ( n < ( 1 + 1 + 1 + 1 )) {
    return leftArea( n, 0, 1 + 1 + 1 + 1, 0.1, 0.8 );
  } else {
	  return topArea( n, 4, 7, 0.1, 0.8 );
  }
}


JFRect JMap2PixelObj::outputArea(int n) {
  return rightArea(n, 0, 1 + 1 + 1);
}


int JMap2PixelObj::inputFace(int n) {
  if ( n < 1 + 1 + 1 + 1) {
    return LEFT;
  } else {
    return TOP;
  }
}


int JMap2PixelObj::outputFace(int n) {
  return RIGHT;
}


JFPoint JMap2PixelObj::inputPoint(int n, JLinkObj& link) {
	if ( n < 4 )
  return leftPoint( n, link, 0, 1 + 1 + 1 + 1, 0.1, 0.8);
	else 
		return topPoint( n, link, 4,7, 0.1, 0.8 );
}


JFPoint JMap2PixelObj::outputPoint(int n, JLinkObj& link) {
  return rightPoint(n, link, 0, 1 + 1 + 1);
}


void JMap2PixelObj::access(int n, JLinkObj& link, const JDataType& data) {
  if ( n == 0 ) {
    INT(data) = in_map;
  } else if ( n == 1 ) {
    INT(data) = pixel_x;
  } else if ( n == 2 ) {
    INT(data) = pixel_y;
  }
}


void JMap2PixelObj::handleInput( int inputDataType, JLinkObj& link, int fieldIndicator ) {
	if ( inputDataType == JIntegerData::id ) {
		int temp;
		link.access( JIntegerData( temp ));
		char dbuf[100];
		sprintf( dbuf, "got int value %d, fieldIndicator %d", temp, fieldIndicator );
		//JComponent::debug( newstrcpy( dbuf ));
		// not sure why I keep getting 0s, every other value is a zero
		if ( temp > 0 ) {
			if ( fieldIndicator == PIXEL_WIDTH ) {
				pixel_width = temp;
			} else if ( fieldIndicator == PIXEL_HEIGHT ) {
				pixel_height = temp;
			}
		} 
	} else if ( inputDataType == JRealData::id ) {
		double temp;
		link.access( JRealData( temp ));
		char js[100];
		sprintf( js, "%f", temp );
		char dbuf[100];
		if ( temp != 0 ) {
			if ( fieldIndicator == UL_LATITUDE ) {
				ul_latitude = JString( newstrcpy( js ));
				sprintf( dbuf, "UL_LATITUDE mapped %f to %s", temp, (char*)ul_latitude );
				JComponent::debug( newstrcpy( dbuf ));
				redoMapping();
			} else if ( fieldIndicator == UL_LONGITUDE ) {
				ul_longitude = JString( newstrcpy( js ));
				sprintf( dbuf, "UL_LONGITUDE mapped %f to %s", temp, (char*)ul_longitude );
				JComponent::debug( newstrcpy( dbuf ));
				redoMapping();
			} else if ( fieldIndicator == LR_LATITUDE ) {
				lr_latitude = JString( newstrcpy( js ));
				sprintf( dbuf, "LR_LATITUDE mapped %f to %s", temp, (char*)lr_latitude );
				JComponent::debug( newstrcpy( dbuf ));
				redoMapping();
			} else if ( fieldIndicator == LR_LONGITUDE ) {
				lr_longitude = JString( newstrcpy( js ));
				sprintf( dbuf, "LR_LONGITUDE mapped %f to %s", temp, (char*)lr_longitude );
				JComponent::debug( newstrcpy( dbuf ));
				redoMapping();
			}
		} 
	} else if ( inputDataType == JStringData::id ) {
		JString temp;
		link.access( JStringData( temp ));
		char dbuf[100];
		sprintf( dbuf, "got String value %s, fieldIndicator %d", (char*)temp, fieldIndicator );
		JComponent::debug( newstrcpy( dbuf ));
	}
}

void JMap2PixelObj::engine(int n, JLinkObj& link) {
  switch (n) {
  case LATITUDE:
    link.access( JStringData( latitude ));
   remap( longitude, latitude );
   break;
  case LONGITUDE:
    link.access( JStringData( longitude ));
	remap( longitude, latitude );
    break;
  case D_LATITUDE:
	link.access( JRealData( d_latitude ));
	d_remap( d_longitude, d_latitude );
	break;
  case D_LONGITUDE:
	link.access( JRealData( d_longitude ));
	d_remap( d_longitude, d_latitude );
	break;
  case MAP_NAME:
	  {
		  JString temp;
		link.access( JStringData( temp ));
		if ( temp.length() > 0 ) {
		  map_name = temp;
		}
	  }
	  break;
  case UL_LATITUDE:
	  JComponent::debug( "engine got UL_LATITUDE" );
	  handleInput( inputType( n ), link, n );
	  break;
  case UL_LONGITUDE:
	  JComponent::debug( "engine got UL_LONGITUDE" );
	  handleInput( inputType( n ), link, n );
	  break;
  case LR_LATITUDE:
	  JComponent::debug( "engine got LR_LATITUDE" );
	  handleInput( inputType( n ), link, n );
	  break;
  case LR_LONGITUDE:
	  JComponent::debug( "engine got LR_LATITUDE" );
	  handleInput( inputType( n ), link, n );
	  break;
  case PIXEL_WIDTH:
  case PIXEL_HEIGHT:
	  JComponent::debug( "got a map configuration input" );
	  handleInput( inputType( n ), link, n );
	  break;
  default:
    break;
  }
}

void JMap2PixelObj::remap( char* longitude, char* latitude ) {
	if (( strlen( longitude ) > 0 ) && ( strlen( latitude ) > 0 )) {
		pixelMapper.map_location( longitude, latitude );
		dobroadcasts();
	}
	lastRemapType = JStringData::id;
}

void JMap2PixelObj::d_remap( double dlongitude, double dlatitude ) {
	if (( dlongitude != 0 ) && ( dlatitude != 0 )) {
		pixelMapper.d_map_location( dlongitude, dlatitude );
		dobroadcasts();
	}
	lastRemapType = JRealData::id;
}

void JMap2PixelObj::redoMapping() {
	if ( lastRemapType == JRealData::id ) {
		d_remap( d_longitude, d_latitude );
	} else if ( lastRemapType == JStringData::id ) {
		remap( longitude, latitude );
	}
}

void JMap2PixelObj::dobroadcasts() {
		if ( pixelMapper.isLocationInMap() ) {
			if ( !in_map ) {
				  in_map = 65535;
				  broadcast( 0 );
			}
			pixel_x = pixelMapper.getPixelX();
			pixel_y = pixelMapper.getPixelY();
			broadcast( 1 );
			broadcast( 2 );
		 } else {
			if ( in_map ) {
				in_map = 0;
				broadcast( 0 );
			}
		}
}


JArray JMap2PixelObj::getProperties() {
  JArray properties;
  properties.append(JStringProperty("ul_latitude", ul_latitude));
  properties.append(JStringProperty("ul_longitude", ul_longitude));
  properties.append(JStringProperty("lr_latitude", lr_latitude));
  properties.append(JStringProperty("lr_longitude", lr_longitude));
  properties.append(JStringProperty("map_name", map_name));
  properties.append(JIntegerProperty("pixel_width", pixel_width, 1, JInteger::MAX_VALUE));
  properties.append(JIntegerProperty("pixel_height", pixel_height, 1, JInteger::MAX_VALUE));
  return properties;
}


boolean JMap2PixelObj::updateProperty(JProperty& prop) {
  if (prop.getName() == JString("ul_latitude")) {
    ul_latitude = ((JStringProperty*)&prop)->value;
  } else if (prop.getName() == JString("ul_longitude")) {
    ul_longitude = ((JStringProperty*)&prop)->value;
  } else if (prop.getName() == JString("lr_latitude")) {
    lr_latitude = ((JStringProperty*)&prop)->value;
  } else if (prop.getName() == JString("lr_longitude")) {
    lr_longitude = ((JStringProperty*)&prop)->value;
  } else if (prop.getName() == JString("map_name")) {
    map_name = ((JStringProperty*)&prop)->value;
  } else if (prop.getName() == JString("pixel_width")) {
    pixel_width = ((JIntegerProperty*)&prop)->value;
  } else if (prop.getName() == JString("pixel_height")) {
    pixel_height = ((JIntegerProperty*)&prop)->value;
  } else {
	return false;
  }
  updatePixelOutputs();
  return true;
}

void JMap2PixelObj::updatePixelOutputs() {
  // pixelMapper constructor values have changed, create a new one 
  // that matches
  if (( ul_longitude.length() > 0 ) && ( lr_longitude.length() > 0 ) &&
	  ( ul_latitude.length() > 0 ) && ( lr_latitude.length() > 0 ) &&
	  ( pixel_width > 1 ) && ( pixel_height > 1 ) &&
	  ( longitude.length() > 0 ) && ( latitude.length() > 0 )) {
	  pixelMapper.init( ul_longitude, lr_longitude, ul_latitude, lr_latitude, map_name,
		  pixel_width, pixel_height );
	  redoMapping();
	  if ( pixelMapper.isLocationInMap() ) {
		  in_map = 65535;
		  broadcast( 0 );
		  pixel_x = pixelMapper.getPixelX();
		  pixel_y = pixelMapper.getPixelY();
		  broadcast( 1 );
		  broadcast( 2 );
	  } else {
		in_map = 0;
		broadcast( 0 );
	  }
  }

}


